"use strict";(self.webpackChunktiled=self.webpackChunktiled||[]).push([[741],{48741:function(e,t,r){r.r(t);var s=r(72791),n=r(94070),a=r(47047),i=r(80184),c=(0,s.lazy)((function(){return Promise.all([r.e(494),r.e(446),r.e(858),r.e(470),r.e(883)]).then(r.bind(r,76883))})),u=(0,s.lazy)((function(){return Promise.all([r.e(494),r.e(858),r.e(321)]).then(r.bind(r,321))}));t.default=function(e){if(e.structure.micro.fields)return(0,i.jsxs)(n.Z,{severity:"warning",children:['This is a "record array" with a'," ",(0,i.jsx)("a",{href:"https://numpy.org/doc/stable/user/basics.rec.html",target:"_blank",rel:"noreferrer",children:"structured data type"}),'. The web interface cannot view it. Use the "Download" tab to access the data.']});switch(e.structure.macro.shape.length<2){case!0:return(0,i.jsx)(s.Suspense,{fallback:(0,i.jsx)(a.Z,{variant:"rectangular"}),children:(0,i.jsx)(c,{segments:e.segments,item:e.item,link:e.link,structure:e.structure})});case!1:return(0,i.jsx)(s.Suspense,{fallback:(0,i.jsx)(a.Z,{variant:"rectangular"}),children:(0,i.jsx)(u,{segments:e.segments,item:e.item,link:e.link,structure:e.structure})})}}}}]);
//# sourceMappingURL=741.fbe96b6d.chunk.js.map