"use strict";(self.webpackChunktiled=self.webpackChunktiled||[]).push([[762],{67762:function(e,t,n){n.r(t);n(72791);var a=n(94070),i=n(80184);t.default=function(e){return(0,i.jsx)(a.Z,{severity:"warning",children:'Bulk download of xarray Datasets is not yet supported by this web interface. Go to the "View" tab and navigate to one the items inside to download.'})}}}]);
//# sourceMappingURL=762.562cdf04.chunk.js.map